/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: c_examine.c,v 1.3 2000/10/09 01:00:10 cosine Exp $
 */

#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "alddefs.h"
#include "defs.h"
#include "load.h"
#include "memory.h"
#include "msg.h"
#include "print.h"

/*
 * libDebug includes
 */
#include "debug.h"

/*
 * libString includes
 */
#include "alloc.h"
#include "Strn.h"

/*
c_examine()
  Examine the debugged process' memory

Format for this command:
  examine [startaddr] [stopaddr] [-num <num>] [-size <val>]
          [-output <letter>]

Options:
  -num <value>         = number of elements to dump
  -size <value/letter> = size of each element
    Possible letter values: b (byte), w (word), h (halfword),
                            d (dword), g (giant - 8 bytes)
  -output <letter>     = output format
    Possible letter values: o (octal), d (decimal), h (hex - default)
                            b (binary)
*/

void
c_examine(int ac, char **av)

{
  int cnt,
      alen;
  char *endptr;
  unsigned long startaddr, /* address to begin memory dump */
                stopaddr;  /* address to stop memory dump */
  int gotaddr;             /* did user supply start address? */
  long elnum,              /* number of elements in memory to dump */
       elsize,             /* size of each element */
       elbytes;            /* number of bytes for all elements */
  unsigned char output;    /* output type (hex, binary, etc) */
  unsigned char *membuf;   /* where to store dumped memory */
  long numbytes;           /* number of bytes of memory to dump */
  long ndumped;            /* number of bytes actually dumped */
  unsigned long deladdr;   /* delta address (stopaddr - startaddr) */

  /*
   * Defaults: dump 20 elements, each 1 byte, in hex format
   */
  elnum = 20;
  elsize = 1;
  output = (unsigned char) 'h';
  startaddr = 0;
  stopaddr = 0;
  gotaddr = 0;

  for (cnt = 1; cnt < ac; ++cnt)
  {
    alen = strlen(av[cnt]);

    if (!Strncasecmp(av[cnt], "-num", alen))
    {
      if (++cnt >= ac)
      {
        Print(P_ERROR, "No element number specified");
        return;
      }

      elnum = strtol(av[cnt], &endptr, 0);
      if ((endptr == av[cnt]) || (*endptr != '\0') || (elnum <= 0))
      {
        Print(P_ERROR, "Invalid element number: %s", av[cnt]);
        return;
      }
    }
    else if (!Strncasecmp(av[cnt], "-size", alen))
    {
      if (++cnt >= ac)
      {
        Print(P_ERROR, "No element size specified");
        return;
      }

      elsize = GetElementSize(av[cnt]);
      if (!elsize)
      {
        Print(P_ERROR, "Invalid element size: %s", av[cnt]);
        return;
      }
    }
    else if (!Strncasecmp(av[cnt], "-output", alen))
    {
      if (++cnt >= ac)
      {
        Print(P_ERROR, "No element size specified");
        return;
      }

      output = (unsigned char) *av[cnt];

      /*
       * We'll let OutputMemory() do any error checking on the
       * output type
       */
    }
    else
    {
      if (!gotaddr)
      {
        gotaddr = 1;
        startaddr = strtol(av[cnt], &endptr, 0);
      }
      else
        stopaddr = strtol(av[cnt], &endptr, 0);

      if ((endptr == av[cnt]) || (*endptr != '\0'))
      {
        Print(P_ERROR, "Invalid address: %s",
          av[cnt]);
        return;
      }
    }
  } /* for (cnt = 1; cnt < ac; ++cnt) */

  if (!gotaddr)
  {
    /*
     * If no start address was specified, the first choice is
     * the address of the next instruction about to be executed.
     * Second choice is the program's entry point, since if
     * NextInstruction is 0, the program hasn't started yet.
     */
    if (NextInstruction)
      startaddr = NextInstruction;
    else
      startaddr = VirtualEntryPoint;
  }

  membuf = 0;

  /*
   * Determine exactly how many bytes of memory we want to
   * dump
   */
  elbytes = elnum * elsize;
  if (stopaddr >= startaddr)
    deladdr = stopaddr - startaddr;
  else
    deladdr = 0;

  if (deladdr > elbytes)
    numbytes = deladdr;
  else
    numbytes = elbytes;

  ndumped = DumpMemory((char **) &membuf, startaddr, numbytes);

  if (membuf)
  {
    if (ndumped > 0)
      OutputMemory(membuf, startaddr, ndumped, elsize, output);
    MyFree(membuf);
  }

  if (ndumped < numbytes)
  {
    Print(P_ERROR, MSG_NOACCESS,
      startaddr + ndumped,
      strerror(errno));
  }
} /* c_examine() */
